from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .middleware import Limiter
from .decorators import limit, exempt, shared_limit
from .errors import RateLimitExceeded




import logging
from django.core.urlresolvers import resolve
from limits.util import parse


def get_ipaddr(request):
    """
    :return: the ip address for the current request (or 127.0.0.1 if none found)
    """
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR', '127.0.0.1')
    return ip


class LimitWrapper(object):
    """
    basic wrapper to encapsulate limits and their context
    """
    def __init__(self, limit, key_func, scope, per_method=False):
        self._limit = limit
        self.key_func = key_func
        self._scope = scope
        self.per_method = per_method

    def get_limit(self, request):
        return parse(self._limit(request)) if callable(self._limit) else self._limit

    def get_scope(self, request):
        return (
            self._scope(resolve(request.path).url_name) if callable(self._scope) else (
                self._scope if self._scope else resolve(request.path).url_name
            )
        )
class BlackHoleHandler(logging.StreamHandler):
    def emit(*_):
        return



from django.http import HttpResponse


class RateLimitExceeded(HttpResponse):
    """
    exception raised when a rate limit is hit (status code: 429).
    """
    status_code = 429
    def __init__(self, limit):
        super(RateLimitExceeded, self).__init__(str(limit))


import importlib
import logging

from django.conf import settings
from django.core.urlresolvers import resolve
from limits.storage import storage_from_string
from limits.strategies import STRATEGIES
from limits.errors import ConfigurationError
from limits.util import parse_many
import six

from .decorators import DECORATED, EXEMPT
from .util import get_ipaddr, LimitWrapper, BlackHoleHandler

from .errors import RateLimitExceeded

class C:
    ENABLED = "RATELIMIT_ENABLED"
    HEADERS_ENABLED = "RATELIMIT_HEADERS_ENABLED"
    STORAGE_URL = "RATELIMIT_STORAGE_URL"
    STRATEGY = "RATELIMIT_STRATEGY"
    GLOBAL_LIMITS = "RATELIMIT_GLOBAL"
    HEADER_LIMIT = "RATELIMIT_HEADER_LIMIT"
    HEADER_REMAINING = "RATELIMIT_HEADER_REMAINING"
    HEADER_RESET = "RATELIMIT_HEADER_RESET"
    DEFAULT_KEY_FUNCTION = "RATELIMIT_KEY_FUNCTION"
    CALLBACK = "RATELIMIT_CALLBACK"

class HEADERS:
    RESET = 1
    REMAINING = 2
    LIMIT = 3


class Limiter(object):
    """

    """
    def __init__(self):
        conf_limits = getattr(settings, C.GLOBAL_LIMITS, "")
        callback = getattr(settings, C.CALLBACK, self.__raise_exceeded )
        self.enabled = getattr(settings, C.ENABLED, True)
        self.headers_enabled = getattr(settings, C.HEADERS_ENABLED, False)
        self.strategy = getattr(settings, C.STRATEGY, 'fixed-window')
        if self.strategy not in STRATEGIES:
            raise ConfigurationError("Invalid rate limiting strategy %s" % self.strategy)
        self.storage = storage_from_string(getattr(settings, C.STORAGE_URL, "memory://"))
        self.limiter = STRATEGIES[self.strategy](self.storage)
        self.key_function = getattr(settings, C.DEFAULT_KEY_FUNCTION, get_ipaddr)
        self.global_limits = []
        if conf_limits:
            self.global_limits = [
                LimitWrapper(
                    limit, self.key_function, None, False
                ) for limit in parse_many(conf_limits)
            ]
        self.header_mapping = {
            HEADERS.RESET : getattr(settings,C.HEADER_RESET, "X-RateLimit-Reset"),
            HEADERS.REMAINING : getattr(settings,C.HEADER_REMAINING, "X-RateLimit-Remaining"),
            HEADERS.LIMIT : getattr(settings,C.HEADER_LIMIT, "X-RateLimit-Limit"),
        }
        self.logger = logging.getLogger("djlimiter")
        self.logger.addHandler(BlackHoleHandler())

        if isinstance(callback, six.string_types):
            mod, _, name = callback.rpartition(".")
            try:
                self.callback = getattr(importlib.import_module(mod), name)
            except AttributeError:
                self.logger.error(
                    "Unable to load callback function %s. Rate limiting disabled" % callback
                )
                self.enabled = False
        else:
            self.callback = callback

    def __raise_exceeded(self, limit):
        return RateLimitExceeded(limit)


    def process_request(self, request):
        """
        :param request:
        :return:
        """
        func = resolve(request.path).func
        name = resolve(request.path).url_name if func else ""
        limits = self.global_limits

        if (
            not self.enabled
            or func in EXEMPT
            or not name
        ):
            return

        if func in DECORATED:
            if func in DECORATED:
                limits = DECORATED[func]
        limit_for_header = None
        failed_limit = None
        for lim in limits:
            limit_scope = lim.get_scope(request) or name
            cur_limit = lim.get_limit(request)
            if not limit_for_header or cur_limit < limit_for_header[0]:
                limit_for_header = (cur_limit, (lim.key_func or self.key_function)(request), limit_scope)
            if lim.per_method:
                limit_scope += ":%s" % request.method
            if not self.limiter.hit(cur_limit, (lim.key_func or self.key_function)(request), limit_scope):
                failed_limit = cur_limit
                limit_for_header = (cur_limit, (lim.key_func or self.key_function)(request), limit_scope)

        request.view_rate_limit = limit_for_header
        if failed_limit:
            return self.callback(failed_limit)

    def process_response(self, request, response):
        """

        :param request:
        :param response:
        :return:
        """
        current_limit = getattr(request, "view_rate_limit", None)
        if self.headers_enabled and current_limit:
            window_stats = self.limiter.get_window_stats(*current_limit)
            response[self.header_mapping[HEADERS.LIMIT]] = str(current_limit[0].amount)
            response[self.header_mapping[HEADERS.REMAINING]] = window_stats[1]
            response[self.header_mapping[HEADERS.RESET]] = window_stats[0]
        return response








from functools import wraps
from limits.util import parse, parse_many
from .util import LimitWrapper

DECORATED = {}
EXEMPT = []

def limit(limit_value, key_function=None, per_method=False):
    """
    decorator to be used for rate limiting individual views

    :param limit_value: rate limit string or a callable that returns a string.
     :ref:`ratelimit-string` for more details.
    :param function key_func: function/lambda to extract the unique identifier for
     the rate limit. defaults to remote address of the request.
    :param bool per_method: whether the limit is sub categorized into the http
     method of the request.
    """
    def __inner(fn):
        @wraps(fn)
        def _inner(*args, **kwargs):
            return fn(*args, **kwargs)
        if callable(limit_value):
            DECORATED.setdefault(_inner, []).append(
                LimitWrapper(limit_value, key_function, None, per_method)
            )
        else:
            DECORATED.setdefault(_inner, []).extend([
                LimitWrapper(
                    lim, key_function, None, per_method
                ) for lim in parse_many(limit_value)
            ])

        return _inner
    return __inner

def shared_limit(limit_value, scope, key_function=None):
    """
    decorator to be applied to multiple views sharing the same rate limit.

    :param limit_value: rate limit string or a callable that returns a string.
     :ref:`ratelimit-string` for more details.
    :param scope: a string or callable that returns a string
     for defining the rate limiting scope.
    :param function key_func: function/lambda to extract the unique identifier for
     the rate limit. defaults to remote address of the request.
    """
    def __inner(fn):
        @wraps(fn)
        def _inner(*args, **kwargs):
            return fn(*args, **kwargs)

        if callable(limit_value):
            DECORATED.setdefault(_inner, []).append(
                LimitWrapper(limit_value, key_function, scope)
            )
        else:
            DECORATED.setdefault(_inner, []).extend([
                LimitWrapper(
                    lim, key_function, scope
                ) for lim in parse_many(limit_value)
            ])
        return _inner
    return __inner



def exempt(fn):
    """
    decorator to mark a view or all views in a blueprint as exempt from rate limits.

    :param fn: the view to wrap.
    :return:
    """
    @wraps(fn)
    def __inner(*a, **k):
        return fn(*a, **k)
    EXEMPT.append(__inner)
    return __inner
